/****************************************************************************
**
** Copyright (C) 2006-2008 fullmetalcoder <fullmetalcoder@hotmail.fr>
**
** This file is part of the Edyuk project <http://edyuk.org>
** 
** This file may be used under the terms of the GNU General Public License
** version 2 as published by the Free Software Foundation and appearing in the
** file GPL.txt included in the packaging of this file.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
****************************************************************************/

#ifndef _AUTOMATON_H_
#define _AUTOMATON_H_

#include <QHash>
#include <QVector>
#include <QDataStream>

typedef quint32 StateId;
typedef QVector<StateId> StateList;

enum TransitionType
{
	CharMask	= 0x0000ffff,
	
	Char		= 0x00000000,
	Epsilon		= 0x01000000,
	Any			= 0x02000000,
	Alias		= 0x04000000,
	
	Increment	= 0x00010000,
	Decrement	= 0x00020000,
	JumpNZ		= 0x00040000,
	JumpZ		= 0x00080000,
	
	TypeMask	= 0xffff0000
};

typedef quint32 Transition;
typedef QVector<Transition> TransitionList;

#define qTransitionType(t) (t & TypeMask)
#define qTransitionChar(t) (t & CharMask)

enum MatchType
{
	Token		= 0x00000000,
	Context		= 0x10000000,
	
	Begin		= 0x01000000,
	End			= 0x02000000,
	
	Indent		= 0x00100000,
	Fold		= 0x00200000,
	Parenthesis	= 0x00400000,
	
	NoConsume	= 0x08000000,
	
	BlackHole	= 0x20000000,
	WormGate	= 0x40000000,
	
	Anchor		= 0x80000000
};

struct Match
{
	inline Match()
	 : type(0), format(0) {}
	
	inline Match(quint32 t, quint32 fmt)
	 : type(t), format(fmt) {}
	
	quint32 type;
	quint32 format;
};

Q_DECLARE_TYPEINFO(Match, Q_MOVABLE_TYPE);

#define qMatchType(m)   ((m).type & 0xfff00000)
#define qMatchJump(m)   ((m).type & 0x000fffff)
#define qMatchFormat(m) ((m).format)

#define qMatch(t, j, f) Match(t | j, f)

enum ContextFlag
{
	None		= 0x0000,
	StayOnLine	= 0x0001,
	Discardable	= 0x0002,
	Internal	= 0x0004,
};

Q_DECLARE_FLAGS(ContextFlags, ContextFlag);
Q_DECLARE_OPERATORS_FOR_FLAGS(ContextFlags);

typedef StateId StateData;

typedef QHash<quint16, StateData> StateHash;

struct HashedAutomaton
{
	bool det;
	QString id;
	quint16 cref;
	quint32 format;
	quint32 flags;
	
	HashedAutomaton *parent;
	
	QVector<StateHash> states;
	
	StateList standalone;
	
	QList<HashedAutomaton*> sub;
	QHash<StateId, Match> match;
	
	HashedAutomaton();
	~HashedAutomaton();
	
	void ref();
	void deref();
	
	void squeeze();
	
	static HashedAutomaton* deserialize(QDataStream& s);
	static void serialize(HashedAutomaton *a, QDataStream& s);
	
	StateId createState();
	StateId createStandaloneState();
	void addTransition(StateId from, Transition t, StateId to);
	
	HashedAutomaton* clone(quint32 formatOffset = 0) const;
};

struct Automaton
{
	bool det;
	QString id;
	quint16 cref;
	quint32 format;
	quint32 flags;
	
	Automaton *parent;
	
	QVector<StateList> edges;
	QVector<TransitionList> trans;
	
	StateList standalone;
	
	QList<Automaton*> sub;
	QHash<StateId, Match> match;
	
	Automaton();
	~Automaton();
	
	void ref();
	void deref();
	
	Automaton* clone(quint32 formatOffset = 0) const;
	
	StateId createState();
	StateId createStandaloneState();
	void addTransition(StateId from, Transition t, StateId to);
	
	StateId addWord(const QString& pattern, bool caseSensitive = true, Match m = qMatch(Token, 0, 0));
	StateId addPattern(const QString& pattern, bool caseSensitive = true, Match m = qMatch(Token, 0, 0));
	
	void addAutomatonContents(Automaton *a, quint32 formatOffset = 0);
	void insertAutomatonContents(Automaton *a, StateId start, StateId end, quint32 formatOffset = 0);
	
	Automaton* determinize() const;
	HashedAutomaton* hashedDeterminize() const;
	
	StateList move(StateList l, Transition t) const;
	StateList epsilonClosure(StateId s) const;
	StateList epsilonClosure(const StateList& l) const;
};

#endif // ! _AUTOMATON_H_
