/****************************************************************************
**
** Copyright (C) 2006-2008 fullmetalcoder <fullmetalcoder@hotmail.fr>
**
** This file is part of the Edyuk project <http://edyuk.org>
** 
** This file may be used under the terms of the GNU General Public License
** version 2 as published by the Free Software Foundation and appearing in the
** file GPL.txt included in the packaging of this file.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
****************************************************************************/

#include "generator.h"

#include "automaton.h"

//#include "lexer.h"
//#include "parser.h"
//#include "action.h"

#include <QHash>
#include <QFile>
#include <QStack>
#include <QTextStream>
#include <QStringList>

static void addPattern(
				Automaton *a,
				const QByteArray& pattern,
				int i, int end,
				StateId forkBeg, StateId forkEnd,
				const Generator::TokenMap& tokens,
				const QHash<QByteArray, Automaton*>& aliases
				);

class ByteArrayRef
{
	public:
		inline ByteArrayRef() : length(0), d(0) {}
		
		inline ByteArrayRef(const QByteArray& b, int pos, int len)
		 : length(len), d(b.constData() + pos)
		{}
		
		inline ByteArrayRef(const ByteArrayRef& b)
		 : length(b.length), d(b.d)
		{}
		
		inline bool isValid() const
		{ return d; }
		
		inline bool isNull() const
		{ return !d; }
		
		inline bool isEmpty() const
		{ return !(d && length); }
		
		inline int size() const
		{ return length; }
		
		inline int count() const
		{ return length; }
		
		ByteArrayRef& operator = (const ByteArrayRef& b)
		{
			d = b.d;
			length = b.length;
			
			return *this;
		}
		
		bool operator != (const QByteArray& b) const
		{
			if ( b.size() != length )
				return true;
			
			return qstrncmp(d, b.constData(), length);
		}
		
		bool operator == (const QByteArray& b) const
		{
			if ( b.size() != length )
				return false;
			
			return !qstrncmp(d, b.constData(), length);
		}
		
		bool operator != (const char *s) const
		{
			int i = 0;
			
			while ( i < length )
			{
				if ( (s[i] != d[i]) || (s[i] == '\0') )
					return true;
				
				++i;
			}
			
			return false;
		}
		
		bool operator == (const char *s) const
		{
			int i = 0;
			
			while ( i < length )
			{
				if ( (s[i] != d[i]) || (s[i] == '\0') )
					return false;
				
				++i;
			}
			
			return true;
		}
		
		inline operator QByteArray() const
		{
			return QByteArray(d, length);
		}
		
		inline operator QString() const
		{
			return QString::fromLocal8Bit(d, length);
		}
		
		inline QByteArray toByteArray() const
		{
			return QByteArray(d, length);
		}
		
		inline QString toString() const
		{
			return QString::fromLocal8Bit(d, length);
		}
		
	private:
		int length;
		const char *d;
};


static QString prettifyTransition(Transition t)
{
	if ( t == Epsilon )
		return QString::fromUtf8("\xcf\xb5");
	else if ( t == Any )
		return "(any)";
	else if ( t == '\n' )
		return QString("\\n");
	else if ( t == '\"' )
		return QString("\\\"");
	
	return QString(QChar(t));
}

static void dump(Automaton *a, QString& dot)
{
	dot += QString("\t// %1 states \n").arg(QString::number(a->trans.count()));
	
	for ( int i = 0; i < a->trans.count(); ++i )
	{
		bool match = a->match.contains(i);
		
		if ( match )
		{
			dot += QString("\tnode [shape = doublecircle]; s%1;\n")
						.arg(QString::number(i));
		}
		
		dot += QString("\t// state %1 : %2 transitions\n").arg(i).arg(a->trans.at(i).count());
		
		for ( int j = 0; j < a->trans.at(i).count(); ++j )
		{
			dot += QString("\ts%1 -> s%2 [label = \"%3\"]\n")
						.arg(QString::number(i))
						.arg(QString::number(a->edges.at(i).at(j)))
						.arg(QString::number(a->trans.at(i).at(j)))
						//.arg(prettifyTransition(a->trans.at(i).at(j)))
						;
			
		}
		
		if ( match )
		{
			Match m = a->match.value(i);
			
			if ( (qMatchType(m) & Context) && (qMatchType(m) & Begin) )
			{
				dot += QString("\tsubgraph cxt_%1 {\n")
							.arg(qMatchJump(m));
				
				dump(a->sub.at(qMatchJump(m)), dot);
				
				dot += "\t}\n";
			}
			
			dot += "\tnode [shape = circle];\n";
		}
	}
}

static void dump(HashedAutomaton *a, QString& dot)
{
	dot += QString("\t// %1 states \n").arg(QString::number(a->states.count()));
	
	for ( int i = 0; i < a->states.count(); ++i )
	{
		bool match = a->match.contains(i);
		
		dot += QString("\ts%1 [shape = %2];\n")
					.arg(QString::number(i))
					.arg(match ? "doublecircle" : "circle");
		
		//dot += QString("\t// state %1 : %2 transitions\n").arg(i).arg(a->trans.at(i).count());
		
		const StateHash& sh = a->states.at(i);
		StateHash::const_iterator it = sh.begin();
		
		while ( it != sh.end() )
		{
			dot += QString("\ts%1 -> s%2 [label = \"%3\"]\n")
						.arg(QString::number(i))
						#ifdef GOOGLE_HASH
						.arg(QString::number(it->second))
						.arg(prettifyTransition(it->first))
						#else
						.arg(QString::number(*it))
						.arg(QString::number(it.key()) ) //prettifyTransition(it.key()))
						#endif
						;
			
			++it;
		}
		
		if ( match )
		{
			Match m = a->match.value(i);
			
			if ( (qMatchType(m) & Context) && (qMatchType(m) & Begin) )
			{
				dot += QString("\tsubgraph cxt_%1 {\n")
							.arg(qMatchJump(m));
				
				//dump(a->sub.at(qMatchJump(m)), dot);
				
				dot += "\t}\n";
			}
		}
	}
}

Generator::Output Generator::create(const QString& lexerDef, const QString& parserDef) const
{
	Output out;
	
	out.lexer = createLexer(lexerDef, out.tokenIdMap);
	out.parser = createParser(parserDef, out.tokenIdMap, out.actions);
	
	#if 1
	{
	QString dot;
	dot +=
		"digraph automaton {\n"
		"\trankdir=LR;\n"
		"\tnode [shape = doublecircle]; s0;\n"
		"\tnode [shape = circle];\n"
		;
	
	::dump(out.lexer, dot);
	
	dot += "}\n";
	
	QFile f("lexer.dot");
	
	if ( f.open(QFile::WriteOnly | QFile::Text) )
		QTextStream(&f) << dot;
	}
	
	{
	QString dot;
	dot +=
		"digraph automaton {\n"
		"\trankdir=LR;\n"
		"\tnode [shape = doublecircle]; s0;\n"
		"\tnode [shape = circle];\n"
		;
	
	::dump(out.parser, dot);
	
	dot += "}\n";
	
	QFile f("parser.dot");
	
	if ( f.open(QFile::WriteOnly | QFile::Text) )
		QTextStream(&f) << dot;
	}
	#endif
	
	return out;
}


HashedAutomaton* Generator::createLexer(const QString& file, Generator::TokenMap& tokens) const
{
	QFile f(file);
	
	if ( !f.open(QFile::ReadOnly | QFile::Text) )
		return 0;
	
	qDebug("generating lexer from : %s", qPrintable(file));
	
	Automaton *a = new Automaton;
	
	bool cs = true;
	bool comment = false;
	int tc = tokens.count(), ln = -1;
	
	while ( !f.atEnd() )
	{
		++ln;
		QByteArray l = f.readLine();
		
		if ( l.at(l.count() - 1) == '\n' )
			l.chop(1);
		
		int i = 0, equ = -1;
		
		if ( comment )
		{
			//qDebug("skipping line %i", ln);
			
			i = l.indexOf("*/");
			
			if ( i == -1 )
				continue;
			
			comment = false;
			i += 2;
		}
		
		QByteArray cleanId;
		QList<int> composite;
		
		while ( (i < l.count()) && equ == -1 )
		{
			quint16 c = l.at(i);
			
			if ( c == '/' )
			{
				++i;
				
				if ( i >= l.count() )
					break;
				
				if ( l.at(i) == '/' )
				{
					//qDebug("skipping line %i", ln);
					break;
				} else if ( l.at(i) == '*' ) {
					comment = true;
					//qDebug("skipping line %i", ln);
					break;
				}
			} else if ( c == '=' ) {
				equ = i;
				break;
			} else if ( c == ':' ) {
				composite << i;
			}
			
			if ( c != '$' && c != '@' )
				cleanId += c;
			
			++i;
		}
		
		if ( equ == -1 )
			continue;
		
		int flagIdx = composite.count() ? composite.last() + 1 : 0;
		
		bool rx = l.at(flagIdx) == '$';
		bool exclu = l.at(rx ? flagIdx + 1 : flagIdx) != '@';
		
		int flagOff = ((rx ? 1 : 0) + (exclu ? 0 : 1));
		
		ByteArrayRef id(l, flagIdx + flagOff, equ - (flagIdx + flagOff) - 1);
		
		++equ;
		
		//qDebug("%s", l.mid(flagIdx + flagOff, equ - (flagIdx + flagOff) - 1).constData());
		
		ByteArrayRef seq(l, equ, l.count() - equ);
		
		Automaton *p = 0, *c = a;
		
		if ( composite.count() )
		{
			// deals with context here
			
			// populate context stack
			QList<ByteArrayRef> contexts;
			
			int last = 0;
			
			for ( i = 0; i < composite.count(); ++i )
			{
				contexts << ByteArrayRef(l, last, composite.at(i) - last);
				last = composite.at(i) + 1;
			}
			
			// create contexts if needed
			
			foreach ( ByteArrayRef r, contexts )
			{
				p = c;
				c = 0;
				
				//qDebug("context search : %s", r.toByteArray().constData());
				
				foreach ( Automaton *child, p->sub )
				{
					if ( child->id == r.toString() )
					{
						c = child;
						break;
					}
				}
				
				if ( !c )
				{
					c = new Automaton;
					c->id = r.toString();
					c->parent = p;
					
					//qDebug("adding context %s to %s", qPrintable(c->id), qPrintable(p->id));
					
					p->sub << c;
				}
			}
			
			// handle flags & special tokens
			
			//qDebug("special token : %s", qPrintable(id.toString()));
			
			if ( id == "start" )
			{
				//qDebug("adding context %s to %s", qPrintable(c->id), qPrintable(p->id));
				
				StateId id;
				
				if ( rx )
					id = p->addPattern(seq.toString(), cs);
				else
					id = p->addWord(seq.toString(), cs);
				
				p->match[id] = qMatch(Context | Begin | (exclu ? Context : NoConsume), p->sub.indexOf(c), -1);
				
				continue;
			} else if ( id == "stop" ) {
				StateId id;
				
				if ( rx )
					id = c->addPattern(seq.toString(), cs);
				else
					id = c->addWord(seq.toString(), cs);
				
				c->match[id] = qMatch(Context | End | (exclu ? Context : NoConsume), 0, -1);
				
				continue;
			} else if ( id == "flag" ) {
				
				if ( seq == "merge" )
				{
					c->flags = 1;//Lexer::Merge;
					tokens[c->id.toLocal8Bit()] = c->format = tc;
					
					++tc;
					
				} else if ( seq == "hide" ) {
					c->flags |= 2; //Lexer::Hide;
				} else {
					qWarning("Unrecognized context flag");
				}
				
				continue;
			}
		}
		
		// deals with "simple" tokens here
		
		// register token id on first occurence, retrieve token id on next
		int tid = 0;
		QByteArray bid = cleanId;
		
		if ( bid.isEmpty() )
			bid = seq.toByteArray();
		
		TokenMap::const_iterator it = tokens.constFind(bid);
		
		if ( it == tokens.constEnd() )
		{
			tid = tc++;
			tokens[bid] = tid;
		} else {
			tid = *it;
		}
		
		//qDebug("adding pattern for : %s [%s]", seq.toByteArray().constData(), bid.constData());
		
		// add token pattern to automaton
		
		int matchFlag = exclu ? Token : NoConsume;
		
		if ( c->flags == 1 ) //Lexer::Merge )
		{
			if ( rx )
			{
				c->addPattern(seq.toString(), cs);
			} else {
				c->addWord(seq.toString(), cs);
				
				tokens[seq.toByteArray()] = tid;
			}
		} else {
			if ( rx )
			{
				c->addPattern(seq.toString(), cs, qMatch(matchFlag, 0, tid));
			} else {
				c->addWord(seq.toString(), cs, qMatch(matchFlag, 0, tid));
				
				tokens[seq.toByteArray()] = tid;
			}
		}
	}
	
	HashedAutomaton *d = a->hashedDeterminize();
	a->deref();
	return d;
}

struct CxtBlob
{
	inline CxtBlob() : state(0), context(0) {}
	inline CxtBlob(Automaton *a) : state(0), context(a) {}
	
	StateId state;
	Automaton *context;
};

HashedAutomaton* Generator::createParser(const QString& file, const Generator::TokenMap& tokens, QList<Action*>& actions) const
{
	QFile f(file);
	
	if ( !f.open(QFile::ReadOnly | QFile::Text) )
		return 0;
	
	qDebug("generating parser from : %s", qPrintable(file));
	
	Automaton *a = new Automaton;
	
	int ln = -1;
	bool comment = false;
	
	QHash<QByteArray, Automaton*> aliases;
	QByteArray lastPattern, lastCxtPattern;
	
	QStack<CxtBlob> blobs;
	blobs.push(a);
	
	while ( !f.atEnd() )
	{
		++ln;
		QByteArray l = f.readLine().simplified();
		
		if ( l.isEmpty() )
			continue;
		
		int i = 0, equ = -1;
		
		if ( comment )
		{
			//qDebug("skipping line %i", ln);
			
			i = l.indexOf("*/");
			
			if ( i == -1 )
				continue;
			
			comment = false;
			i += 2;
			
			if ( i >= l.count() )
				continue;
		}
		
		CxtBlob& blob = blobs.top();
		
		quint16 c = l.at(i);
		
		if ( c == '/' )
		{
			// comment line
			
			++i;
			
			if ( i >= l.count() )
				continue;
			
			if ( l.at(i) == '/' )
			{
				//qDebug("skipping line %i", ln);
				continue;
			} else if ( l.at(i) == '*' ) {
				comment = true;
				//qDebug("skipping line %i", ln);
				continue;
			} else {
				qWarning("syntax error at line %i", ln);
				continue;
			}
		} else if ( c == '$' ) {
			// alias
			
			int equ = l.indexOf('=');
			
			QByteArray alias = l.mid(1, equ - 1).trimmed();
			
			Automaton *sa = new Automaton;
			sa->id = alias;
			
			StateId end = sa->createState();
			addPattern(sa, l, equ + 1, l.count(), 0, end, tokens, aliases);
			sa->match[end] = qMatch(Token, 0, 0);
			
			aliases[alias] = sa;
			
		} else if ( c == '@' ) {
			// macro
			
			ByteArrayRef macro(l, 1, l.count() - 1);
			
			if ( blob.context && blobs.count() > 1 )
			{
				if ( macro == "import" )
				{
					//qDebug("importing (recurse on)");
					blob.context->addAutomatonContents(blobs.at(blobs.count() - 2).context);
				} else if ( macro == "recursive" ) {
					
					//qDebug("recursing");
					StateId end = blob.context->createState();
					addPattern(blob.context, lastCxtPattern, 0, lastCxtPattern.count(), 0, end, tokens, aliases);
					
					blob.context->match.insertMulti(blob.state, qMatch(Context | Begin, blob.context->sub.count(), 0));
					blob.context->sub << blob.context;
				}
			}
			
		} else if ( c == '>' ) {
			// rule
			
			if ( blob.context && blob.state )
			{
				// TODO : parse rule, create proper action and set format accordingly
				//qDebug("match(0x%x, %i)", blob.context, blob.state);
				blob.context->match.insertMulti(blob.state, qMatch(Token, 0, actions.count()));
				
				//actions << Action::build(l, 1);
			}
		} else if ( c == '{' ) {
			// context enter
			
			lastCxtPattern = lastPattern;
			
			if ( blob.context && blob.state )
			{
				//qDebug("context enter.");
				Automaton *child = new Automaton;
				child->parent = blob.context;
				
				blob.context->match.insertMulti(blob.state, qMatch(Context | Begin, blob.context->sub.count(), 0));
				blob.context->sub << child;
				
				blobs.push(child);
			} else {
				qDebug("broken...");
				blobs.push(0);
			}
		} else if ( c == '}' ) {
			// context leave
			
			if ( blob.context && blob.state )
			{
				//qDebug("context exit.");
				blob.context->match.insertMulti(blob.state, qMatch(Context | End, 0, 0));
			}
			
			blobs.pop();
		} else {
			// expression
			
			lastPattern = l;
			
			if ( blob.context )
			{
				//qDebug("pattern %s", l.constData());
				
				StateId end = blob.context->createState();
				addPattern(blob.context, l, 0, l.count(), 0, end, tokens, aliases);
				
				blob.state = end;
			}
		}
	}
	
	HashedAutomaton *d = a->hashedDeterminize();
	a->deref();
	return d;
}

bool checkConsumeFlag(int& i, const QString& pattern, int idx)
{
	while ( i < idx && pattern.at(i).isSpace() )
	{
		++i;
		
		if ( pattern.at(i) == '@' )
		{
			++i;
			qDebug("detected NoConsume tag : %s", qPrintable(pattern.mid(i, idx - i)));
			return true;
		}
	}
	
	return false;
}

void addPattern(
				Automaton *a,
				const QByteArray& pattern,
				int i, int end,
				StateId forkBeg, StateId forkEnd,
				const Generator::TokenMap& tokens,
				const QHash<QByteArray, Automaton*>& aliases
				)
{
	if ( !a )
		return;
	
	if ( end < 0 )
		end = pattern.length() + end + 1;
	else
		end = qMin(end, pattern.length());
	
	//qDebug("addPattern(\"%s\");", pattern.mid(i, end - i).constData());
	
	--i;
	
	StateId previous = forkBeg, current = forkBeg, fork = 0xffffffff;
	
	while ( ++i < end )
	{
		ByteArrayRef terminal;
		char c = pattern.at(i);
		
		if ( isspace(c) )
			continue;
		
		if ( c == '(' )
		{
			int idx = i, nest = 1;
			
			while ( nest && (++idx < end) )
			{
				if ( pattern.at(idx) == '\\' )
				{
					++idx;
					continue;
				} else if ( pattern.at(idx) == '(' ) {
					++nest;
				} else if ( pattern.at(idx) == ')' ) {
					--nest;
				} else if ( pattern.at(idx) == '[' ) {
					while ( ++idx < end )
					{
						if ( pattern.at(idx) == '\\' )
						{
							++idx;
							continue;
						} else if ( pattern.at(idx) == ']' ) {
							break;
						}
					}
				}
			}
			
			if ( (idx >= end) || nest )
			{
				qWarning("Parsing error : supplied regexp is syntactically invalid.\n[%s]\nIrregularity from pos %i.", pattern.constData(), i);
				return;
			}
			
			StateId sbeg = a->createState(),
					send = a->createState();
			
			a->addTransition(current, Epsilon, sbeg);
			
			addPattern(a, pattern, i + 1, idx, sbeg, send, tokens, aliases);
			
			previous = current;
			current = send;
			fork = sbeg;
			i = idx;
			
			continue;
			
		} else if ( c == '{' || c == '[' || c == '<' || c == '/' ) {
			// new operator : internal context with(out) recursion ('<' : voided)...
			
			bool quote = false;
			int idx = i, nest = 1, blocknest = 0;
			
			QList<int> portions;
			
			while ( nest && (++idx < end) )
			{
				char cc = pattern.at(idx);
				
				if ( cc == '\'' )
				{
					quote = !quote;
				} else if ( quote ) {
					if ( cc == '\\' )
						++idx;
					
				} else {
					if ( cc == '{' || cc == '[' || cc == '<' || cc == '/' )
					{
						++nest;
					} else if ( cc == '}' || cc == ']' || cc == '>' || cc == '\\' ) {
						--nest;
					} else if ( cc == '(' ) {
						++blocknest;
					} else if ( cc == ')' ) {
						--blocknest;
					} else if ( nest == 1 && cc == '.' ) {
						portions << idx;
						
						//qDebug("portion at : %i", idx);
					}
				}
			}
			
			if ( ((idx >= end) && nest) || portions.isEmpty()  )
			{
				qWarning("Parsing error : supplied regexp is syntactically invalid.\n[%s]\nIrregularity from pos %i.", pattern.constData(), i);
				qDebug("idx = %i, nest = %i, portions.count() = %i", idx, nest, portions.count());
				return;
			}
			
			StateId sbeg = a->createState(),
					send = a->createState();
			
			a->addTransition(current, Epsilon, sbeg);
			
			++i;
			bool doNotConsume = checkConsumeFlag(i, pattern, idx);
			
			addPattern(a, pattern, i, portions.at(0), sbeg, send, tokens, aliases);
			
			// add match for context begin
			a->match[send] = qMatch(Context | Begin | BlackHole | (doNotConsume ? NoConsume : 0), a->sub.count(), 0);
			
			// add new context
			Automaton *sub = new Automaton;
			sub->parent = a;
			
			if ( c == '<' || c == '/' )
			{
				a->match[sbeg] = qMatch(Anchor, a->sub.count(), 0);
				sub->flags = Discardable;
			}
			
			a->sub << sub;
			
			bool recursive = c == '{' || c == '<';
			
			if ( recursive )
			{
				// recursion trick
				StateId lrec = sub->createState();
				addPattern(sub, pattern, i, portions.at(0), 0, lrec, tokens, aliases);
				sub->match[lrec] = qMatch(Context | Begin | BlackHole | (doNotConsume ? NoConsume : 0), sub->sub.count(), 0);
				sub->sub << sub;
			}
			
			// add context end
			StateId lend = sub->createState();
			
			int tmpIdx = portions.at(0) + 1;
			int max = portions.count() > 1 ? portions.at(1) : idx;
			doNotConsume = checkConsumeFlag(tmpIdx, pattern, idx);
			
			addPattern(sub, pattern, tmpIdx, max, 0, lend, tokens, aliases);
			sub->match[lend] = qMatch(Context | End | BlackHole | (doNotConsume ? NoConsume : 0), a->standalone.count(), 0);
			
			if ( recursive )
			{
				// now this is really a shitty hack but we gotta have the same number of
				// standalone states or recursive matching will simply fail...
				while ( sub->standalone.count() < a->standalone.count() )
					sub->standalone << 0xffffffff;
				
				// now the magic : just make sure you go back to the start state after a recursion
				sub->standalone << 0;
			}
			
			if ( portions.count() > 1 )
			{
				// optional parts
				
				if ( c == '{' || c == '<' )
				{
					// add emergency exit
					
					tmpIdx = portions.at(1) + 1;
					max = portions.count() > 2 ? portions.at(2) : idx;
					doNotConsume = checkConsumeFlag(tmpIdx, pattern, max);
					
					if ( tmpIdx < max )
					{
						StateId lexit = sub->createState();
						addPattern(sub, pattern, tmpIdx, max, 0, lexit, tokens, aliases);
						sub->match[lexit] = qMatch(Context | End | WormGate | (doNotConsume ? NoConsume : 0), 0, 0);
					}
					
					if ( portions.count() > 2 )
					{
						addPattern(sub, pattern, portions.at(2) + 1, portions.count() > 3 ? portions.at(3) : idx, 0, sub->createState(), tokens, aliases);
					}
				} else {
					// add internal matches
					
					addPattern(sub, pattern, portions.at(1) + 1, portions.count() > 2 ? portions.at(2) : idx, 0, sub->createState(), tokens, aliases);
				}
			}
			
			//qDebug("internal cxt %i", a->standalone.count());
			StateId rets = a->createStandaloneState();
			
			previous = current;
			current = rets;
			fork = sbeg;
			i = idx;
			
			continue;
			
		} else if ( c == '|' ) {
			//fork = 0xffffffff;
			
			a->addTransition(current, Epsilon, forkEnd);
			
			fork = previous = current = forkBeg;
			
			continue;
		} else if ( c == '*' ) {
			
			//a->addTransition(current, Epsilon, (fork == 0xffffffff) ? current : fork);
			a->addTransition(current, Epsilon, fork);
			a->addTransition(fork, Epsilon, current);
			
			//StateId next = a->createState();
			
			//a->addTransition(current, Epsilon, next);
			//a->addTransition((fork == 0xffffffff) ? previous : fork, Epsilon, next);
			//current = next;
			
			continue;
		} else if ( c == '+' ) {
			// simple auto loop...
			
			//a->addTransition(current, Epsilon, (fork == 0xffffffff) ? current : fork);
			a->addTransition(current, Epsilon, fork);
			
			continue;
		} else if ( c == '?' ) {
			
			//StateId next = a->createState();
			
			//a->addTransition(current, Epsilon, next);
			//a->addTransition((fork == 0xffffffff) ? previous : fork, Epsilon, next);
			//current = next;
			
			a->addTransition(fork, Epsilon, current);
			
			continue;
		} else if ( c == '\'' || c == '\"' ) {
			// litteral terminals
			
			int len = 0, idx = i;
			
			while ( ++idx < end )
			{
				if ( pattern.at(idx) == c )
					break;
				
				++len;
				//a->addTransition(sbeg, pattern.at(i).unicode(), send);
			}
			
			if ( idx >= end )
			{
				i = idx;
				qWarning("missing end quote for terminal.");
				continue;
			}
			
			terminal = ByteArrayRef(pattern, i + 1, len);
			i = idx;
			
		} else if ( c == '$' ) {
			
			int len = 0, idx = i;
			
			while ( ++idx < end )
			{
				char cc = pattern.at(idx);
				
				if ( !isalnum(cc) && (cc != '_') )
					break;
				
				++len;
				//a->addTransition(sbeg, pattern.at(i).unicode(), send);
			}
			
			//terminal = ByteArrayRef(pattern, i, len);
			
			// add an alias...
			QByteArray alias = pattern.mid(i + 1, len);
			QHash<QByteArray, Automaton*>::const_iterator it = aliases.constFind(alias);
			
			if ( it == aliases.constEnd() )
			{
				qWarning("Reference to undeclared alias : \'%s\'", alias.constData());
			} else {
				previous = current;
				
				fork = a->createState();
				a->addTransition(current, Epsilon, fork);
				
				current = a->createState();
				
				a->insertAutomatonContents(*it, fork, current);
			}
			
			i = idx - 1;
			continue;
		} else {
			
			int len = 1, idx = i;
			
			while ( ++idx < end )
			{
				char cc = pattern.at(idx);
				
				if ( !isalnum(cc) && (cc != '_') )
					break;
				
				++len;
				//a->addTransition(sbeg, pattern.at(i).unicode(), send);
			}
			
			terminal = ByteArrayRef(pattern, i, len);
			i = idx - 1;
		}
		
		if ( terminal.isNull() )
			continue;
		
		Generator::TokenMap::const_iterator it = tokens.constFind(terminal.toByteArray());
		
		if ( it == tokens.constEnd() )
		{
			qWarning("Reference to undeclared token : \'%s\'", terminal.toByteArray().constData());
			continue;
		}
		
		previous = current;
		
		fork = a->createState();
		a->addTransition(current, Epsilon, fork);
		
		current = a->createState();
		a->addTransition(fork, *it, current);
	}
	
	a->addTransition(current, Epsilon, forkEnd);
}
